import argparse
import collections
import datetime
import json
import os.path
import subprocess
import sys
import tempfile
import urllib
import urlparse

from pkg_resources import resource_filename
from update_package.common import is_url

def load_project(project):
    """
    Load project description from file
    """
    if '/' not in project:
        project = resource_filename(__name__, project)
    with open(project) as fileh:
        jdata = json.load(fileh)
    return jdata


def checkout_repo(repo):
    """
    Checkout git repository
    """
    without_ext, _ = os.path.splitext(repo)
    repo_dir = os.path.basename(without_ext)

    if os.path.exists(repo_dir):
        subprocess.check_call(['git', 'pull', '-r'], cwd=repo_dir)
    else:
        subprocess.check_call(['git', 'clone', repo])
    requests_dir = os.path.join(repo_dir, 'requests')
    if not os.path.exists(requests_dir):
        os.mkdir(requests_dir)

    return repo_dir


def get_url(name, build_url):
    """
    Return a URL for the deliverable
    """
    if is_url(name):
        return name
    else:
        return os.path.join(build_url, name)


def get_deliverables_url(branch, brand=None, build=None, base_url=None):
    """
    Return the URL to query the deliverables service
    """
    default_base_url = \
        "http://util.xenrt.citrite.net/web-scripts/cgi/xs/deliverable.cgi"
    base_url = base_url or default_base_url

    params = {
        "branch" : branch,
        "deliverable" : "release"
    }    

    if build:
        params["build"] = build
    if brand:
        params["brand"] = brand
        
    url_parts = urlparse.urlparse(base_url)
    query = urlparse.parse_qs(url_parts.query)
    query.update(params)
    replaced_url_parts = url_parts._replace(query=urllib.urlencode(query))
    url = urlparse.urlunparse(replaced_url_parts)

    return url



class Deliverable_Type:
    """
    A simple class of const values
    """
    UPDATE = "updates",
    SOURCE = "sources",
    COPY_ONLY = "copy_only"


"""
A Deliverable is a source or update to be signed, or a file or directory to be
copied into the release
type: One of the Deliverable_Type consts
name: The name of the deliverable, eg. ack_iso or xencenter.iso
url: The absolute path to this deliverable
build_dir: Path to its build dir, which is copied into the release (optional)
"""
Deliverable = collections.namedtuple("Deliverable", ["type", "name", "url",
                                                     "build_dir"])


def get_remote_deliverables(config, brand):
    """
    Get the deliverables from the deliverables service
    """
    # Config may either be a branch name, or an object containing branch
    # and build.
    if isinstance(config, basestring):
        deliverables_url = get_deliverables_url(config, brand)
    else:
        branch = config.get('branch')
        build = config.get('build', None)
        base_url = config.get('base_url', None)
        deliverables_url = get_deliverables_url(branch, brand, build, base_url)

    response = urllib.urlopen(deliverables_url)
    data = json.loads(response.read()) 

    release_data = [(k,v) for k,v in data.iteritems() if v['release']]

    read_deliverables = []
    
    for (name, val) in release_data:
        build_dir = val.get('build_dir', None)

        type = Deliverable_Type.COPY_ONLY
        if "sign" in val:
            # Convert to our expected type to avoid unicode/string issues
            if val["sign"] == "updates":
                type = Deliverable_Type.UPDATE
            elif val["sign"] == "sources":
                type = Deliverable_Type.SOURCE       
 
        d = Deliverable(type, name, val["url"], build_dir)
        read_deliverables.append(d)
        
    return read_deliverables


def get_project_deliverables(build_url, updates, sources, copy_urls):
    """
    Get the deliverables from the project file
    """
    project_deliverables = []
    
    # An update/source can be either a file name, or an object containing name
    # and build_dir
    for update in updates:
        if isinstance(update, basestring):
            name = update
            build_dir = None
        else:
            name = update["name"]
            build_dir = update.get("build_dir", None)
        
        url = get_url(name, build_url)

        d = Deliverable(Deliverable_Type.UPDATE, name, url, build_dir)
        project_deliverables.append(d)

    for source in sources:
        if isinstance(source, basestring):
            name = source
            build_dir = None
        else:
            name = source["name"]
            build_dir = source.get("build_dir", None)
        
        url = get_url(name, build_url)

        d = Deliverable(Deliverable_Type.SOURCE, name, url, build_dir)
        project_deliverables.append(d)

    for to_copy in copy_urls:
        url = get_url(to_copy, build_url)
        name = os.path.basename(url)
        d = Deliverable(Deliverable_Type.COPY_ONLY, name, url, None)
        project_deliverables.append(d)

    return project_deliverables


def get_duplicates(xs):
    if len(xs) == len(set(xs)):
        return None
    dupes = [x for x, count in collections.Counter(xs).items() if count > 1]
    return dupes

def check_merge_deliverables(project_deliverables, remote_deliverables):
    """
    Merge the two deliverables lists, failing if there are any duplicate names
    """
    # Check no same name deliverables
    project_names = [d.name for d in project_deliverables]
    remote_names = [d.name for d in remote_deliverables]   
 
    project_duplicates = get_duplicates(project_names)
    if project_duplicates:
        sys.exit("FATAL: Duplicate name found in project file \n " \
                 "%s" % project_duplicates)

    # This is unlikely, but best to check just in case - otherwise we'd get a
    # confusing error
    remote_duplicates = get_duplicates(remote_names)
    if remote_duplicates: 
        sys.exit("FATAL: Duplicate name found in deliverables service \n " \
                 " %s" % remote_duplicates)

    merged_names = project_names + remote_names
    merged_duplicates = get_duplicates(merged_names)
    if merged_duplicates:
        sys.exit("FATAL: Duplicate name found in both project file and remote " \
              "deliverables \n %s" % merged_duplicates)

    return project_deliverables + remote_deliverables

    
def build_request(args, project):
    """
    Build request from deliverables service, project file and command line
    """

    remote_deliverables = []
    if 'deliverables' in project:
        deliverables = project['deliverables']
        brand = project.get('brand', None)
        remote_deliverables = get_remote_deliverables(deliverables, brand)

    updates = []
    if 'updates' in project:
        updates = project['updates']
    if args.update:
        updates += args.update

    sources = []
    if 'sources' in project:
        sources = project['sources']
    if args.source:
        sources += args.source

    copy_urls = []
    if 'copy_files' in project:
        copy_urls = project['copy_files']

    project_deliverables = get_project_deliverables(args.build_url, updates,
                                                    sources, copy_urls)

    merged_deliverables = check_merge_deliverables(project_deliverables,
                                                   remote_deliverables)
    
    request = {}
    request['release'] = args.release_url
    request['key'] = project['key']
    request['keyfiles'] = project['keyfiles']

    if 'media' in project:
        media = get_url(project['media'], args.build_url)
        
        request['media'] = media

    request['updates'] = [d.url for d in merged_deliverables if \
                                d.type == Deliverable_Type.UPDATE]
    request['sources'] = [d.url for d in merged_deliverables if \
                                d.type == Deliverable_Type.SOURCE] 
    request['copy_files'] = [d.url for d in merged_deliverables if \
                                d.type == Deliverable_Type.COPY_ONLY]

    copy_dirs = []
    if project.get('copy_assembly', False):
        copy_dirs.append({
	    'url' : args.build_url,
	    'release_path' : 'assembly'
	})

    build_dir_prefix = project.get('build_dir_prefix', None)    
 
    for d in merged_deliverables:
        if not d.build_dir:
            continue
        
        if not build_dir_prefix:
            print("WARN: build_dir was ignored for deliverable %s, set " \
                  "build_dir_prefix in project file to copy the build_dir." %
                  d.name)
            continue

        release_build_dir_name = build_dir_prefix + d.name
        copy_dirs.append({
            'url' : d.build_dir,
            'release_path' : release_build_dir_name
        })

    if copy_dirs:
        request['copy_dirs'] = copy_dirs

    copy_dirs = []
    if project.get('copy_assembly', False):
        copy_dirs.append({
	    'url' : args.build_url,
	    'release_path' : 'assembly'
	})

    if copy_dirs:
        request['copy_dirs'] = copy_dirs

    if args.dry_run:
        print "DRY RUN. Would submit request:"
        print json.dumps(request, sort_keys=True, indent=4,
                         separators=(',', ': '))

    return request


def add_request(args, repo_dir, project):
    """
    Build request, sign it and add it locally
    """
    request = build_request(args, project)
    if args.dry_run:
        return
    datestamp = datetime.datetime.utcnow().isoformat() + '-'
    request_dir = os.path.join(repo_dir, 'requests')
    with tempfile.NamedTemporaryFile(prefix=datestamp, suffix='.json',
                                     dir=request_dir,
                                     delete=False) as request_file:
        request_name = request_file.name
        request_signed = request_name + ".signed"
        print >> request_file, json.dumps(request, sort_keys=True, indent=4,
                                          separators=(',', ': '))
    cmd = ['gpg', '--output', request_signed]
    if args.local_user:
        cmd += ['--local-user', args.local_user]
    cmd += ['--sign', request_name]
    subprocess.check_call(cmd)
    cmd = ['git', 'add', request_signed]
    subprocess.check_call(cmd, cwd=repo_dir)
    cmd = ['git', 'commit', '-m', "Added " + os.path.basename(request_signed)]
    subprocess.check_call(cmd, cwd=repo_dir)


def push_request(repo_dir):
    """
    Submit request to repository
    """
    subprocess.check_call(['git', 'pull', '-r'], cwd=repo_dir)
    subprocess.check_call(['git', 'push'], cwd=repo_dir)


def parse_args_or_exit(argv=None):
    """
    Parse command line options
    """
    parser = argparse.ArgumentParser(description="Generate build signing request")
    parser.add_argument("build_url",
                        help="Artifactory URL of build to sign")
    parser.add_argument("release_url",
                        help="Artifactory URL to upload to")
    parser.add_argument("--repo",
                        default="ssh://git@code.citrite.net/xsc/xs-sign.git",
                        help="git repository submit request to")
    parser.add_argument("--project", required=True,
                        help="Project description")
    parser.add_argument("--local-user", "--default-key", "-u")
    parser.add_argument("--update", action="append",
                        help="Artifactory URL of update")
    parser.add_argument("--source", action="append",
                        help="Artifactory URL of source")
    parser.add_argument("--dry-run", action="store_true", 
                        help="Print request and exit without submitting")

    return parser.parse_args(argv)


def main(argv):
    """
    Create a build signing request
    """
    args = parse_args_or_exit(argv)
    print args

    project = load_project(args.project)
    repo_dir = checkout_repo(args.repo)
    add_request(args, repo_dir, project)
    if not args.dry_run:
        push_request(repo_dir)

def _main():
    """
    Entry point for setuptools CLI wrapper
    """
    main(sys.argv[1:])


# Entry point when run directly
if __name__ == "__main__":
    _main()
