#!/usr/env python

import argparse
import json
import os
import subprocess
import sys
import urllib2
import base64
import getpass
import urlparse
from pkg_resources import resource_filename

vendor = {"ICS": "Inspur", "XS": "Citrix"}

def get_auth(title="Username: "):
    u = raw_input(title)
    p = getpass.getpass("[%s] password: " % u)
    return "%s:%s" % (u, p)

def get_json_data(url, auth=None):
    req = urllib2.Request(url)
    if auth:
        auth_string = base64.b64encode(auth)
        req.add_header("Authorization", "Basic %s" % auth_string)
    response = urllib2.urlopen(req)
    return json.loads(response.read())

def get_jira_data(update_ticket):
    """
    Get JIRA data for the ticket
    """
    jira_url = "https://issues.citrite.net/rest/api/2/issue/%s" % update_ticket
    return get_json_data(jira_url, auth=get_auth(title='JIRA user: '))

def get_version_data(version_url):
    return get_json_data(version_url)

def get_assembly_url(jira_data):
    """
    Extract the assembly URL field
    """
    field_name = "customfield_20941" # Hardcoded for now
    return jira_data["fields"][field_name]    


def get_version(jira_data, version_data):
    """ 
    Extract the fix version in major.minor.patch form
    """

    fix_version = jira_data["fields"]["fixVersions"][0] # List, just get the first
    fix_version_name = fix_version['name'] # eg. XS 7.1 (Ely)
    vendor_prefix = fix_version_name.split(' ')[0] # ICS, XS
    fix_version_rel_name = fix_version_name.split('(')[1]
    fix_version_rel_name = fix_version_rel_name[:-1]
    return version_data[vendor[vendor_prefix]]["Versions"][fix_version_rel_name]

def get_rc_url(dev_signed_url, version):
    """
    Generate the release-candidate URL
    """
    parsed = urlparse.urlsplit(dev_signed_url)
    path = parsed.path
    path_parts = path.split('/')
    end_parts = path_parts[-3:] # Last 3 parts are UPD-<num>/<build>/<filename>
    upd_part, build_num, filename = end_parts
    
    major, _  = version.split(".", 1)
    version_major_part = "XenServer-%s.x" % major
    version_minor_part = "XS-" + version
    version_parts = [version_major_part, version_minor_part] 

    name, _ = os.path.splitext(filename)
    name_upd_part = "%s-%s" % (name, upd_part)

    new_path_parts = ["xs-local-release-candidate"] + version_parts + \
                     ["updates", name_upd_part, build_num, filename]

    new_path ='/'.join(new_path_parts)
    replaced = parsed._replace(path=new_path)
    
    return urlparse.urlunsplit(replaced)


def get_command(project, dev_url, signed_url, dry_run=False):
    """
    Generate the request-sign-build command
    """
    update_file = os.path.basename(dev_url)
    
    base_filename, update_extension = os.path.splitext(update_file)
    
    sources_file = "%s-sources%s" % (base_filename, update_extension)
    binpkgs_file = "%s-binpkgs%s" % (base_filename, update_extension)
    
    source_path = os.path.dirname(dev_url)
    dest_path = os.path.dirname(signed_url)

    cmd = ["request-sign-build", "--project", project, "--update", update_file, 
           "--source", sources_file, "--source", binpkgs_file, source_path, 
           dest_path]
    
    if dry_run:
        cmd.append("--dry-run")

    return cmd


def parse_args_or_exit(argv=None):
    """
    Parse command line options
    """
    parser = argparse.ArgumentParser(description="Wrapper for signing hotfixes")

    parser.add_argument("--project", required=True,
                        help="Project description")
    parser.add_argument("--ticket", required=True, 
                        help="JIRA ticket. eg. UPD-52")
    parser.add_argument("--dry-run", action="store_true",
                        help="Include dry-run in the output command")
    parser.add_argument("--print-command", action="store_true",
                        help="Print the command that would run and then exit " \
                             "without running it")    
    # Once product-version.json is hosted on web, default_ver_url needs to
    # point to the correct url.
    default_ver_url = "file://" + resource_filename(__name__,
                                    'product-version.json')
    parser.add_argument("--version-url", dest='verurl',
                        default=default_ver_url, help="Version URL")

    return parser.parse_args(argv)


def main(argv):
    """
    Generate a request-sign-build call
    """
    args = parse_args_or_exit(argv)

    jira_data = get_jira_data(args.ticket)
    version_data = get_version_data(args.verurl)
    assembly_url = get_assembly_url(jira_data)
    fix_version = get_version(jira_data, version_data)
    rc_url = get_rc_url(assembly_url, fix_version)
    cmd = get_command(args.project, assembly_url, rc_url, args.dry_run)

    if args.print_command:
        print "Would run:"
        print ' '.join(cmd)
    else:
        subprocess.check_call(cmd)


def _main():
    """
    Entry point for setuptools CLI wrapper
    """
    main(sys.argv[1:])


# Entry point when run directly
if __name__ == "__main__":
    _main()
    
